//#include "client.hpp"
//#include "server.hpp"
#include "interactive.hpp"
#include "blc_program.h"
#include "blc_channel.h"
#include "core.hpp"

#include <sys/errno.h>
#include <sys/mman.h>//shm_unlink

static void unlink(char const*unlink_name, char const* force){
    blc_channel channel_info;
    
    if (blc_channel_get_info_with_name(&channel_info, unlink_name) != -1){
        channel_info.remove();
    }
    else{
        color_fprintf(BLC_RED, stderr, "Channel '%s' does not exist.\n", unlink_name);
        if (force){
            fprintf(stderr, "We try to unlink the shared memory\n");
            if (shm_unlink(unlink_name) == -1) color_fprintf(BLC_RED, stderr, "Impossible to remove shared memory '%s'.\n%s\n", unlink_name, strerror(errno));
            else color_fprintf(BLC_GREEN, stderr, "'%s' has been removed.", unlink_name);
        }
        else fprintf(stderr, "You can try to remove the shared memory anyway using the --force option.\n");
    }
}

static void unlink_unused(char const *start_filter_name){
    int channels_nb;
    blc_channel *tmp_channels;
    
    channels_nb = blc_channel_get_all_infos_with_filter(&tmp_channels, start_filter_name);
    channels.assign(tmp_channels, tmp_channels+channels_nb);
    unlink_unused();
    FREE(tmp_channels); //This is redundant because mix C-array / vector
    channels.clear();
}

/**This main function parses the parameters and calls either the command line mode or the interactive mode.*/
int main(int argc, char **argv){
    char const  *unlink_name, *unlink_unused_flag, *force, *start_filter_name, *server_mode, *client_mode, *port_str, *without_processes;
    
    blc_program_set_description("Analyses and manages blc channels");
    blc_program_add_option(&client_mode, 'c', "client", "server ip", "run as client", nullptr);
    blc_program_add_option(&server_mode, 's', "server", nullptr, "run as server", nullptr);
    blc_program_add_option(&port_str, 'p', "port", "integer", "port of the server", "55555");
    blc_program_add_option(&unlink_name, 'u', "unlink", "blc_channel", "unlink the channel", NULL);
    blc_program_add_option(&without_processes, 'w', "without-processes", nullptr, "do not display processes ( much faster on OSX)", nullptr);
    //  blc_program_add_option(&uncontrol_id, 'C', "uncontrol", "id", "unlink the semaphores", NULL);
    blc_program_add_option(&unlink_unused_flag, 'U', "unlink_unused", NULL, "unlink unused channels", NULL);
    blc_program_add_option(&force, 'f', "force", NULL, "force the action", NULL);
    blc_program_add_parameter(&start_filter_name, "name_filter", 0, "filter the beginning of the blc_channel name", NULL);
    blc_program_init(&argc, &argv, nullptr);
    
    if (unlink_name) unlink(unlink_name, force);
    else if (unlink_unused_flag) unlink_unused(start_filter_name);
    else {
        atexit(blc_quit);
  //      if(server_mode && client_mode) EXIT_ON_ERROR("You cannot request client and server");
    //    if (server_mode) start_server(port_str);
     //   else if (client_mode) start_client(client_mode, port_str);
//        else 
            main_interactive_mode(start_filter_name, without_processes==nullptr);
    }
    return EXIT_SUCCESS;
}
    
