#include "client.hpp"
#include "core.hpp"  //get_channel_index_by_name_or_id

#include "connexion.hpp"
#include "pthread.h"
#include "interactive.hpp"

#include "blc_core.h"
#include "blc_realtime.h"
#include "blc_network.h"
#include "blc_channel.h"
#include "blc_program.h"
#include <vector>
#include <deque>

#include <string.h> //strsep

using namespace std;

vector <blc_channel*> local_channels_pt; //Much simpller than an unordered_map

struct Remote:Connexion{
   vector<blc_channel*> remote_channels_pt; // This associate id with pointer to channel (simpler and faster thant unordered map
    deque<blc_channel> remote_channels; //This store the content of the channel desciption
    Remote(blc_network *network);
    
    void read_channel_list();
    void read_channel_data();
    blc_channel *get_channel_by_name_or_id(char const *name_or_id);
};

Remote::Remote(blc_network *network):Connexion(network) {
}

blc_channel *Remote::get_channel_by_name_or_id(char const *name)
{
    int id;
    char *end;
    
    if (strchr("/:^.", name[0])){
        for(auto &channel:remote_channels) //it is a par key/value
            if (strcmp(channel.name+1, name+1) == 0) return &channel;
    }
    else {
        id = strtol(name, &end, 10);
        if (end != name && end[0] == 0){
            return &remote_channels[id];
        }
        else color_fprintf(BLC_RED, stderr, "You have to give either an id or a channel name stating by '/' instead of '%s'.\n", name);
    }
    return nullptr;
}

void Remote::read_channel_list(){
    blc_channel *channel;
    channel_id_t channel_id;
    
    fprintf(stderr, "\n@1:%s\n", address_name);
    while (read_pos < network->received_size){
        READ_DATA(channel_id, network->chars, read_pos);

        if (channel_id>=(int)remote_channels_pt.size()) {
            remote_channels_pt.reserve(channel_id);
            channel=nullptr;
        }
        else channel=remote_channels_pt[channel_id];

        if (channel==nullptr){
            remote_channels.emplace_back();
            channel=&remote_channels.back();
            remote_channels_pt[channel_id]=channel;
        }
        
        read_pos+=channel->sscan_info(network->chars + read_pos, true);
        strncat(channel->name, "@1",  sizeof(channel->name)); // latter @1 would be generique with many different server possiblesl
        STRCPY(channel->parameter, address_name);
        channel->fprint_info(stderr, true);
    }
}

void Remote::read_channel_data(){
    blc_channel *local_channel;
    int ret;
    channel_id_t distant_channel_id;
    size_t tmp_read_size, total_read;
    
    READ_DATA(distant_channel_id, network->chars, read_pos);
    
    if (distant_channel_id < (int)local_channels_pt.size()){
        local_channel=local_channels_pt[distant_channel_id];
    }
    else local_channel=nullptr;
    if (local_channel==nullptr){
        blc_channel *distant_channel;
        distant_channel=&remote_channels[distant_channel_id];
        ret=distant_channel->conflict(distant_channel->name);
        
        if (ret==-1) channels.emplace_back(distant_channel->name, BLC_CHANNEL_WRITE, distant_channel->type, distant_channel->format, distant_channel->dims_nb, distant_channel->dims);
        else if (ret==0) channels.emplace_back(distant_channel->name, BLC_CHANNEL_WRITE);
        else EXIT_ON_CHANNEL_ERROR(distant_channel, " conflict '%d'", ret);
        local_channel=&channels.back();
        if (distant_channel_id >= (int)local_channels_pt.size()) local_channels_pt.resize(distant_channel_id+1);
        local_channels_pt[distant_channel_id]=local_channel;
    }
    
    total_read=0;
    
    if (local_channel->sem_ack_data) sem_wait(local_channel->sem_ack_data);
    
    while(1){
        tmp_read_size=network->received_size-read_pos;
        memcpy(local_channel->chars+total_read, network->chars+read_pos,  tmp_read_size);
        read_pos+=tmp_read_size;
        total_read+=tmp_read_size;
        if (message_mode==short_message || message_mode==long_message_end) break;
       
        if ( message_mode==long_message_continue || message_mode==long_message_begin) {
            network->recv_block();
            read_header();
        }
    }
    if (total_read < local_channel->size) EXIT_ON_ERROR("Read only '%lu' bytes  instead of '%lu' bytes", total_read, local_channel->size);
    if (local_channel->sem_new_data) sem_post(local_channel->sem_new_data);
}

void list_distant_channels(char*, void *user_data){
    Connexion *connexion=(Connexion*)user_data;
    connexion->start_data(REQUEST_CHANNEL_LIST);
    connexion->end_data();
}

void refresh_distant_channel(char *name_or_id, void *user_data){
    blc_channel *distant_channel;
    char const* period_str;
    int32_t period_us=-1; // Default refresh value only once
    Remote *remote=(Remote*)user_data;
    
    period_str=strchr(name_or_id, ',');
    if (period_str)  if (sscanf(period_str+1, "%d", &period_us)!=1){ //+1 to avoid first comma
        color_fprintf(BLC_RED, stderr, "Wrong period format in '%s'", period_str+1);
        return;
    }
    distant_channel = remote->get_channel_by_name_or_id(name_or_id);
    if (distant_channel==nullptr) {
        fprintf(stderr, "'%s' is not a valid name or index.", name_or_id);
        return;
    }
    
    remote->start_data(REQUEST_CHANNEL_DATA);
    WRITE_DATA(remote->network->chars, remote->write_pos, distant_channel->id);
    WRITE_DATA(remote->network->chars, remote->write_pos, period_us);
    remote->end_data();
}

void *client_manager(void *user_data){
    Remote *remote=(Remote*)user_data;
    uint8_t command;
    
    while(1){
        remote->network->recv_block();
        remote->read_header();
        READ_DATA(command, remote->network->chars, remote->read_pos);
        switch (command){
            case CHANNEL_DESCRIPTION:
                remote->read_channel_list();
                break;
            case CHANNEL_DATA:
                remote->read_channel_data();
                break;
            default:
                EXIT_ON_ERROR("Unknonw command '%d'", command);
        }
    }
    return nullptr;
}

void start_client(char const *address_name, char const*port_name){
    
    blc_network network(address_name, port_name, BLC_UDP4, -1);
    Remote remote(&network);
    pthread_t thread;
    
    blc_command_add("l", list_distant_channels, nullptr, "Refresh the list of distants channels", &remote);
    blc_command_add("r", refresh_distant_channel, "remote channel[, refresh period(µs)]", "Refresh content of the remote channel", &remote);
    
    BLC_PTHREAD_CHECK(pthread_create(&thread, nullptr, client_manager, &remote), nullptr);
    list_distant_channels(nullptr, &remote);
    
    BLC_COMMAND_LOOP(-1){
        blc_command_display_help();
    }
    BLC_PTHREAD_CHECK(pthread_cancel(thread), nullptr); //If we not stop it and join it, it will crash because the connexion will be closed at the exit of the functionl
    BLC_PTHREAD_CHECK(pthread_join(thread, nullptr), nullptr);
    
}
