#include "connexion.hpp"
#include "blc_network.h"
#include "blc_channel.h"
#include "blc_program.h"

#include <vector>
#include <deque>
#include <map>

#include <mutex>

static const int UDP_IPV4_header_size=28; //20 IPv4 + 8 UDP

using namespace std;


Connexion::Connexion(blc_network *network):network(network){
    remote_address_length=network->remote_address_length;
    remote_address=network->remote_address;
    network->get_remote_address(address_name, sizeof(address_name), port_name, sizeof(port_name));
    mtu=1500; //This is the most common for most interfaces
}

void Connexion::bufferize_data(char const *buffer, size_t data_to_write){
    size_t free_buffer_space;
    size_t written=0;
    
    while(written < data_to_write){
        free_buffer_space=mtu-UDP_IPV4_header_size-write_pos;
        if (data_to_write-written <= free_buffer_space){
            memcpy(network->chars+write_pos, buffer+written, data_to_write-written);
            write_pos+=data_to_write-written;
            written+=data_to_write-written;
        }
        else {
            if (message_mode == short_message) message_mode = long_message_begin;
            else message_mode=long_message_continue;
            memcpy(network->chars+write_pos, buffer+written, free_buffer_space);
            write_pos+=free_buffer_space;
            written+=free_buffer_space;
            flush();
        }
    }
    fprintf(stderr, "Sent '%lu'\n", written);
}

void Connexion::start_data(uint8_t command){
    mutex.lock();
    write_pos=sizeof(message_id_t)+sizeof(message_mode_t); //reserve for message_id and message_mode
    WRITE_DATA(network->chars, write_pos, command);
    message_mode=short_message; //By default we suppose it is a short message
}

void Connexion::first_read_header(){
    message_id_t distant_message_id;
    
    read_pos=0;
    READ_DATA(distant_message_id, network->chars, read_pos);
    receive_message_id=distant_message_id+1;
    READ_DATA(message_mode, network->chars, read_pos);
    fprintf(stderr,"Message mode '%d'\n", message_mode);
}

void Connexion::read_header(){
    message_id_t distant_message_id;
    
    read_pos=0;
    READ_DATA(distant_message_id, network->chars, read_pos);
    if (distant_message_id!=receive_message_id) EXIT_ON_ERROR("expected message id '%d' different from received message id '%d'", distant_message_id, receive_message_id);
    else receive_message_id++;
    READ_DATA(message_mode, network->chars, read_pos);
}


void Connexion::end_data(){
    if (message_mode==long_message_continue) message_mode=long_message_end;
    flush();
    mutex.unlock();

}

void Connexion::flush(){
    memcpy(network->chars, &send_message_id, sizeof(send_message_id));
    memcpy(network->chars+sizeof(send_message_id), &message_mode, sizeof(message_mode));
    network->send(write_pos);
    write_pos=sizeof(send_message_id)+sizeof(message_mode);
    send_message_id++;
}

bool compare_blc_network(blc_network const &network1, blc_network const &network2){ //Used for map
    return (memcmp(&network1.remote_address, &network2.remote_address, network1.remote_address_length)<0);
}

