#ifndef CONNEXION_HPP
#define CONNEXION_HPP

#include "blc_network.h"
#include "blc_channel.h"
#include <inttypes.h>
#include <deque>
#include <mutex>

///Read data with the good size of variable and update the offset accordingly.
#define READ_DATA(variable, pointer, offset) do{memcpy(&variable, pointer+offset, sizeof(variable)); offset+=sizeof(variable);}while(0)
///Writes data with the good size of variable and update the offset accordingly.
#define WRITE_DATA(pointer, offset, variable) do{ memcpy(pointer+offset, &variable, sizeof(variable)); offset+=sizeof(variable);}while(0)

enum {REQUEST_CHANNEL_LIST=0, CHANNEL_DESCRIPTION, REQUEST_CHANNEL_DATA, CHANNEL_DATA};
typedef int32_t channel_id_t, period_us_t; //This ensures that server and client use the same type


struct Connexion{
    typedef uint16_t message_id_t;
    typedef uint8_t message_mode_t;
    std::mutex mutex;
    
    message_mode_t message_mode;
    message_id_t send_message_id=0, receive_message_id=0; //Global like that server and clients use the same
    size_t read_pos, write_pos;
    char address_name[NAME_MAX];
    char port_name[NAME_MAX];
    struct sockaddr_storage remote_address;
    socklen_t remote_address_length;
    blc_network *network;
    int mtu;
    
    static const int long_message_begin=0, long_message_end=1, long_message_continue=2, short_message=3;

    ///Create a connexion based on the network. The main goal of a connextion is to manage a offset (pos) on the current data being bufferized.
    Connexion(blc_network *network);
    
    ///Bufferise th edata in the buffer, if it is too big for a UDP socket, the message is partially send.
    void bufferize_data(char const *buffer, size_t data_size);

    ///Read the header. Checks the message_id and return the command
    void read_header();
    ///Read the header for the first time. The message_id is read anly but not checked. It is supposed to be the first connexion. Returns the command
    void first_read_header();
    ///Bufferize message_id and command
    void start_data(uint8_t command);
    ///Send the end of the buffer
    void end_data();
    void flush();
    
};
#endif
